import os
import json
import pytest
from unittest.mock import MagicMock, patch, mock_open
from usdm3_excel import USDM3Excel
from usdm4_excel.export.base.ct_version import CTVersion
from usdm4_excel.export.excel_table_writer.excel_table_writer import ExcelTableWriter


class TestUSDM3Excel:
    """Tests for the USDM3Excel class."""

    @pytest.mark.xfail(reason="Excel file writing issues in test environment")
    def test_to_excel(self):
        """Test the to_excel method."""
        # Create a USDM3Excel instance
        usdm3_excel = USDM3Excel()

        # Mock the open function
        mock_json_data = {
            "study": {
                "id": "550e8400-e29b-41d4-a716-446655440000",  # Using a valid UUID
                "name": "Test Study",
            },
            "usdmVersion": "3.0.0",
            "instanceType": "Wrapper",
        }
        m = mock_open(read_data=json.dumps(mock_json_data))

        # Create a mock for the ExcelTableWriter
        mock_etw = MagicMock(spec=ExcelTableWriter)

        # Create a mock for the Workbook
        mock_workbook = MagicMock()
        mock_ws = MagicMock()
        mock_workbook.active = mock_ws

        # Mock the necessary classes and methods
        with (
            patch("builtins.open", m),
            patch(
                "os.path.exists", return_value=False
            ),  # Make sure it creates a new workbook
            patch("openpyxl.Workbook", return_value=mock_workbook),
            patch(
                "usdm4_excel.export.base.ct_version.CTVersion"
            ) as mock_ct_version_class,
            patch(
                "usdm4_excel.export.excel_table_writer.excel_table_writer.ExcelTableWriter",
                return_value=mock_etw,
            ),
            patch("usdm4.USDM4") as mock_usdm4_class,
            patch("usdm3_excel.StudySheet") as mock_study_sheet_class,
            patch(
                "usdm3_excel.StudyIdentifiersSheet"
            ) as mock_study_identifiers_sheet_class,
            patch("usdm3_excel.StudyContentSheet") as mock_study_content_sheet_class,
            patch(
                "usdm3_excel.StudyActivitiesSheet"
            ) as mock_study_activities_sheet_class,
            patch("usdm3_excel.StudyTimingSheet") as mock_study_timing_sheet_class,
            patch(
                "usdm3_excel.StudyEncountersSheet"
            ) as mock_study_encounters_sheet_class,
            patch("usdm3_excel.StudyEpochsSheet") as mock_study_epochs_sheet_class,
            patch("usdm3_excel.StudyArmsSheet") as mock_study_arms_sheet_class,
            patch("usdm3_excel.StudyDesignSheet") as mock_study_design_sheet_class,
            patch("usdm3_excel.StudyTimelineSheet") as mock_study_timeline_sheet_class,
            patch("usdm3_excel.ConfigurationSheet") as mock_configuration_sheet_class,
            patch(
                "usdm4_excel.export.study_procedures_sheet.study_procedures_sheet.StudyProceduresSheet"
            ) as mock_study_procedures_sheet_class,
        ):
            # Configure the mocks
            mock_ct_version = MagicMock(spec=CTVersion)
            mock_ct_version_class.return_value = mock_ct_version

            mock_usdm4 = MagicMock()
            mock_wrapper = MagicMock()
            mock_study = MagicMock()
            mock_wrapper.study = mock_study
            mock_usdm4.from_json.return_value = mock_wrapper
            mock_usdm4_class.return_value = mock_usdm4

            # Configure the sheet mocks
            mock_sheets = [
                mock_study_sheet_class.return_value,
                mock_study_identifiers_sheet_class.return_value,
                mock_study_content_sheet_class.return_value,
                mock_study_activities_sheet_class.return_value,
                mock_study_timing_sheet_class.return_value,
                mock_study_encounters_sheet_class.return_value,
                mock_study_epochs_sheet_class.return_value,
                mock_study_arms_sheet_class.return_value,
                mock_study_design_sheet_class.return_value,
                mock_study_timeline_sheet_class.return_value,
                mock_study_procedures_sheet_class.return_value,
                mock_configuration_sheet_class.return_value,
            ]

            # Call the to_excel method
            output_file = "tests/test_files/test.xlsx"
            usdm3_excel.to_excel("test.json", output_file)

            # Verify that the file was opened
            m.assert_called_once_with("test.json")

            # Verify that the CTVersion and ExcelTableWriter were created
            mock_ct_version_class.assert_called_once()

            # Verify that USDM4 was created and from_json was called
            mock_usdm4_class.assert_called_once()
            mock_usdm4.from_json.assert_called_once_with(mock_json_data)

            # Verify that each sheet was created and save was called
            for sheet in mock_sheets:
                sheet.save.assert_called_once_with(mock_study)

            # Verify that the ExcelTableWriter's save method was called
            mock_etw.save.assert_called_once()

            # Clean up
            try:
                os.remove(output_file)
                print(f"File {output_file} deleted")
            except Exception as e:
                print(f"Exception raised delting file {e}")



from usdm4.api.study import Study
from usdm4.api.identifier import StudyIdentifier
from usdm4.api.address import Address
from usdm4.api.organization import Organization
from usdm4.api.study_version import StudyVersion
from usdm4_excel.export.base.collection_panel import CollectionPanel


class IdentifiersPanel(CollectionPanel):
    def execute(self, study: Study) -> list[list[dict]]:
        collection = []
        for version in study.versions:
            for item in version.studyIdentifiers:
                self._add_identifier(collection, item, version)
        return super().execute(
            collection,
            [
                "organizationIdentifierScheme",
                "organizationIdentifier",
                "organizationName",
                "organizationType",
                "studyIdentifier",
                "organizationAddress",
            ],
        )

    def _add_identifier(
        self, collection: list, item: StudyIdentifier, version: StudyVersion
    ):
        org: Organization = version.organization(item.scopeId)
        data = org.model_dump()
        data["organizationIdentifierScheme"] = data["identifierScheme"]
        data["organizationIdentifier"] = data["identifier"]
        data["organizationName"] = data["name"]
        data["organizationType"] = self._map_org_type(self._pt_from_code(org.type))
        data["organizationAddress"] = self._from_address(org.legalAddress)
        data["studyIdentifier"] = item.text
        collection.append(data)

    def _from_address(self, address: Address):
        if address is None:
            return "|||||"
        lines_to_line = (", ").join(address.lines)
        items = [lines_to_line if lines_to_line else ""]
        self._append_address_item(items, address.district)
        self._append_address_item(items, address.city)
        self._append_address_item(items, address.state)
        self._append_address_item(items, address.postalCode)
        code = address.country.code if address.country else ""
        self._append_address_item(items, code)
        return ("|").join(items)

    def _append_address_item(self, items: list[str], value: str) -> None:
        items.append(value if value else "")

    def _map_org_type(self, code: str) -> str:
        code_map = {"Pharmaceutical Company": "Clinical Study Sponsor"}
        return code_map[code] if code in code_map else code



from .identifiers_panel import IdentifiersPanel
from usdm4.api.study import Study
from usdm4_excel.export.base.base_sheet import BaseSheet


class StudyIdentifiersSheet(BaseSheet):
    SHEET_NAME = "studyIdentifiers"

    def save(self, study: Study):
        op = IdentifiersPanel(self.ct_version)
        result = op.execute(study)
        self.etw.add_table(result, self.SHEET_NAME)
        self.etw.format_cells(
            self.SHEET_NAME,
            (1, 1, 1, 6),
            font_style="bold",
            background_color=self.HEADING_BG,
        )
        self.etw.set_column_width(self.SHEET_NAME, [1, 6], 20.0)



from usdm3_excel import USDM3Excel
from tests.helpers.excel_yaml_helper import ExcelYamlHelper

SAVE = False


def run_test(usdm_file, excel_file, yaml_file):
    usdm3_excel = USDM3Excel()
    usdm3_excel.to_excel(usdm_file, excel_file)
    helper = ExcelYamlHelper(excel_file, yaml_file)
    if SAVE:
        helper.save()
    assert helper.compare


def test_integration_1():
    run_test(
        "tests/test_files/usdm_1.json",
        "tests/test_files/usdm_excel_1.xlsx",
        "tests/test_files/usdm_excel_1.yaml",
    )


def test_integration_2():
    run_test(
        "tests/test_files/usdm_2.json",
        "tests/test_files/usdm_excel_2.xlsx",
        "tests/test_files/usdm_excel_2.yaml",
    )



from usdm4.api.study import Study
from usdm4_excel.export.base.collection_panel import CollectionPanel


class StudyPopulationPanel(CollectionPanel):
    def execute(self, study: Study) -> list[list[dict]]:
        collection = []
        self._add_default(collection)
        return super().execute(
            collection,
            [
                "level",
                "name",
                "description",
                "label",
                "plannedCompletionNumber",
                "plannedEnrollmentNumber",
                "plannedAge",
                "plannedSexOfParticipants",
                "includesHealthySubjects",
            ],
        )

    def _add_default(self, collection: list):
        data = {
            "level": "Main",
            "name": "POP1",
            "description": "Default Population",
            "label": "Default Population",
            "plannedCompletionNumber": "0",
            "plannedEnrollmentNumber": "0",
            "plannedAge": "18..100 years",
            "plannedSexOfParticipants": "BOTH",
            "includesHealthySubjects": "N",
        }
        collection.append(data)



import os
from simple_error_log import Errors
from usdm3_excel.export.study_sheet.study_sheet import StudySheet
from usdm3_excel.export.study_population_sheet.study_population_sheet import (
    StudyPopulationSheet,
)
from usdm3_excel.export.study_identifiers_sheet.study_identifiers_sheet import (
    StudyIdentifiersSheet,
)
from usdm3_excel.export.study_content_sheet.study_content_sheet import (
    StudyContentSheet,
)
from usdm4_excel.export.study_activities_sheet.study_activities_sheet import (
    StudyActivitiesSheet,
)
from usdm4_excel.export.study_encounters_sheet.study_encounters_sheet import (
    StudyEncountersSheet,
)
from usdm4_excel.export.study_epochs_sheet.study_epochs_sheet import (
    StudyEpochsSheet,
)
from usdm4_excel.export.study_arms_sheet.study_arms_sheet import (
    StudyArmsSheet,
)
from usdm3_excel.export.study_design_sheet.study_design_sheet import (
    StudyDesignSheet,
)
from usdm3_excel.export.study_timeline_sheet.study_timeline_sheet import (
    StudyTimelineSheet,
)
from usdm4_excel.export.study_procedures_sheet.study_procedures_sheet import (
    StudyProceduresSheet,
)
from usdm4_excel.export.configuration_sheet.configuration_sheet import (
    ConfigurationSheet,
)
from usdm3_excel.export.study_timing_sheet.study_timing_sheet import StudyTimingSheet
from usdm4_excel.export.base.empty_sheet import EmptySheet
from usdm4_excel.export.base.ct_version import CTVersion
from usdm4_excel.export.excel_table_writer.excel_table_writer import ExcelTableWriter
from usdm4 import USDM4
from usdm4.api.wrapper import Wrapper


class USDM3Excel:
    def to_excel(self, usdm_filepath: str, excel_filepath: str):
        ct_version = CTVersion()
        self._remove_exisitng_file(excel_filepath)
        etw = ExcelTableWriter(excel_filepath, default_sheet_name="study")
        usdm = USDM4()
        errors = Errors()
        wrapper: Wrapper = usdm.load(usdm_filepath, errors)
        study = wrapper.study
        empty_sheets = {
            "studyDesignElements": [
                "name",
                "description",
                "label",
                "transitionStartRule",
                "transitionEndRule",
            ],
            "studyDesignIndications": ["name", "description", "label", "codes"],
            "studyDesignInterventions": [
                "name",
                "description",
                "label",
                "codes",
                "role",
                "type",
                "pharmacologicalClass",
                "productDesignation",
                "minimumResponseDuration",
                "administrationName",
                "administrationDescription",
                "administrationLabel",
                "administrationRoute",
                "administrationDose",
                "administrationFrequency",
                "administrationDurationDescription",
                "administrationDurationWillVary",
                "administrationDurationWillVaryReason",
                "administrationDurationQuantity",
            ],
            "studyDesignPopulations": [
                "level",
                "name",
                "description",
                "label",
                "plannedCompletionNumber",
                "plannedEnrollmentNumber",
                "plannedAge",
                "plannedSexOfParticipants",
                "includesHealthySubjects",
            ],
            "studyDesignEstimands": [
                "xref",
                "summaryMeasure",
                "populationDescription",
                "intercurrentEventName",
                "intercurrentEventDescription",
                "intercurrentEventStrategy",
                "treatmentXref",
                "endpointXref",
            ],
            "studyDesignOE": [
                "objectiveName",
                "objectiveDescription",
                "objectiveLabel",
                "objectiveText",
                "objectiveLevel",
                "endpointName",
                "endpointDescription",
                "endpointLabel",
                "endpointText",
                "endpointPurpose",
                "endpointLevel",
            ],
            "studyDesignEligibilityCriteria": [
                "category",
                "identifier",
                "name",
                "description",
                "label",
                "text",
                "dictionary",
            ],
        }
        for sheet_name, column_names in empty_sheets.items():
            _ = EmptySheet(ct_version, etw).blank(column_names, sheet_name)

        for klass in [
            StudySheet,
            StudyPopulationSheet,
            StudyIdentifiersSheet,
            StudyContentSheet,
            StudyActivitiesSheet,
            StudyTimingSheet,
            StudyEncountersSheet,
            StudyEpochsSheet,
            StudyArmsSheet,
            StudyDesignSheet,
            StudyTimelineSheet,
            StudyProceduresSheet,
            ConfigurationSheet,
        ]:
            klass(ct_version, etw).save(study)
        etw.save()

    def _remove_exisitng_file(self, excel_filepath: str) -> None:
        try:
            os.remove(excel_filepath)
        except Exception:
            pass



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

package_info = {}
with open("src/usdm3_excel/__info__.py") as fp:
    exec(fp.read(), package_info)

setuptools.setup(
    name="usdm3_excel",
    version=package_info["__package_version__"],
    author="D Iberson-Hurst",
    author_email="",
    description="A python package for importing and exporting the CDISC TransCelerate USDM, version 3, using Excel",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=["usdm4>=0.12.0", "usdm4_excel>=0.8.1"],
    packages=setuptools.find_packages(where="src"),
    package_dir={"": "src"},
    package_data={},
    tests_require=[
        "pytest",
        "pytest-cov",
        "pytest-mock",
        "python-dotenv",
        "pandas",
        "ruff",
    ],
    classifiers=[
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Operating System :: OS Independent",
    ],
)



from usdm3_excel.__info__ import __package_version__, __model_version__


class TestVersion:
    """Tests for the version module."""

    def test_package_version(self):
        """Test that the package version is correctly defined."""
        # Verify that the version is a string
        assert isinstance(__package_version__, str)

        # Verify that the version follows semantic versioning (x.y.z)
        parts = __package_version__.split(".")
        assert len(parts) == 3
        assert all(part.isdigit() for part in parts)

    def test_model_version(self):
        """Test that the model version is correctly defined."""
        # Verify that the version is a string
        assert isinstance(__model_version__, str)

        # Verify that the version follows semantic versioning (x.y.z)
        parts = __model_version__.split(".")
        assert len(parts) == 3
        assert all(part.isdigit() for part in parts)

    def test_version_matches_package(self):
        """Test that the version matches the package version."""
        # This test just verifies that the version is defined
        from usdm3_excel.__info__ import __package_version__

        # Verify that the version is a string
        assert isinstance(__package_version__, str)

        # Verify that the version follows semantic versioning (x.y.z)
        parts = __package_version__.split(".")
        assert len(parts) == 3
        assert all(part.isdigit() for part in parts)



import re
from usdm4.api.study import Study
from usdm4.api.timing import Timing
from usdm4.api.code import Code
from usdm4.api.schedule_timeline import ScheduleTimeline
from usdm4_excel.export.base.collection_panel import CollectionPanel


class TimingPanel(CollectionPanel):
    def execute(self, study: Study) -> list[list[dict]]:
        collection = []
        for version in study.versions:
            for design in version.studyDesigns:
                for timeline in design.scheduleTimelines:
                    for item in timeline.timings:
                        self._add_timing(collection, item, timeline)
        return super().execute(
            collection,
            [
                "name",
                "description",
                "label",
                "type",
                "from",
                "to",
                "timingValue",
                "toFrom",
                "window",
            ],
        )

    def _add_timing(self, collection: list, item: Timing, timeline: ScheduleTimeline):
        data = item.model_dump()
        data["type"] = self._encode_type(item.type)
        from_tp = timeline.find_timepoint(item.relativeFromScheduledInstanceId)
        data["from"] = from_tp.name if from_tp else ""
        to_tp = timeline.find_timepoint(item.relativeToScheduledInstanceId)
        data["to"] = to_tp.name if to_tp else ""
        data["timingValue"] = self._decode_iso8601_duration(item.value)
        data["window"] = item.windowLabel
        data["toFrom"] = self._encode_to_from(item.relativeToFrom)
        collection.append(data)

    def _encode_type(self, code: Code):
        mapping = {"C201358": "FIXED", "C201356": "AFTER", "C201357": "BEFORE"}
        return mapping[code.code]

    def _encode_to_from(self, code: Code):
        mapping = {
            "C201355": "S2S",
            "C201354": "S2E",
            "C201353": "E2S",
            "C201352": "E2E",
        }
        return mapping[code.code]

    def _decode_iso8601_duration(self, value: str) -> str:
        p_units_map = {
            "Y": "Years",
            "M": "Months",
            "W": "Weeks",
            "D": "Days",
        }
        pt_units_map = {
            "H": "Hours",
            "M": "Minutes",
            "S": "Seconds",
        }
        units_str = None
        units_char = value[-1]
        p_type = value[0:2]
        if p_type.upper() == "PT":
            if units_char in pt_units_map:
                units_str = pt_units_map[units_char]
        else:
            if units_char in p_units_map:
                units_str = p_units_map[units_char]
        if units_str:
            match = re.search(r"\d+", value)
            if match:
                return f"{match.group()} {units_str}"
        return "1 Day"
